sched: Allow more than one running scheduler instance.
[paraslash.git] / sched.c
diff --git a/sched.c b/sched.c
index 7b5ca63d29dc78aeda551bd11fb63819dce7bbad..041cf417723bd848ea56d2c7eb4935943abeeaef 100644 (file)
--- a/sched.c
+++ b/sched.c
@@ -9,6 +9,7 @@
 #include <regex.h>
 #include <assert.h>
 #include <sys/time.h>
+#include <stdbool.h>
 
 #include "para.h"
 #include "ipc.h"
@@ -19,9 +20,6 @@
 #include "time.h"
 #include "error.h"
 
-static struct list_head pre_select_list, post_select_list;
-static int initialized;
-
 static struct timeval now_struct;
 struct timeval *now = &now_struct;
 
@@ -35,8 +33,6 @@ struct timeval *now = &now_struct;
  */
 static void unregister_task(struct task *t)
 {
-       if (!initialized)
-               return;
        assert(t->error < 0);
        PARA_INFO_LOG("unregistering %s (%s)\n", t->status,
                para_strerror(-t->error));
@@ -46,14 +42,25 @@ static void unregister_task(struct task *t)
                list_del(&t->post_select_node);
 }
 
+static inline bool timeout_is_zero(struct sched *s)
+{
+       struct timeval *tv = &s->select_timeout;
+       return tv->tv_sec == 0 && tv->tv_usec == 0;
+}
+
 static void sched_preselect(struct sched *s)
 {
        struct task *t, *tmp;
-       list_for_each_entry_safe(t, tmp, &pre_select_list, pre_select_node) {
-               if (t->pre_select)
-                       t->pre_select(s, t);
-               if (t->error < 0)
+       list_for_each_entry_safe(t, tmp, &s->pre_select_list, pre_select_node) {
+               if (t->error < 0) {
                        unregister_task(t);
+                       continue;
+               }
+               if (!t->pre_select)
+                       continue;
+               t->pre_select(s, t);
+               if (timeout_is_zero(s))
+                       break;
        }
 }
 
@@ -81,7 +88,7 @@ static void sched_post_select(struct sched *s)
 {
        struct task *t, *tmp;
 
-       list_for_each_entry_safe(t, tmp, &post_select_list, post_select_node) {
+       list_for_each_entry_safe(t, tmp, &s->post_select_list, post_select_node) {
                if (t->error >= 0)
                        call_post_select(s, t);
 //             PARA_INFO_LOG("%s: %d\n", t->status, t->ret);
@@ -93,7 +100,7 @@ static void sched_post_select(struct sched *s)
                 * this case t has been unregistered already, so we must not
                 * unregister it again.
                 */
-               if (list_empty(&post_select_list))
+               if (list_empty(&s->post_select_list))
                        return;
                unregister_task(t);
        }
@@ -118,8 +125,6 @@ int schedule(struct sched *s)
 {
        int ret;
 
-       if (!initialized)
-               return -E_NOT_INITIALIZED;
        if (!s->select_function)
                s->select_function = para_select;
 again:
@@ -129,62 +134,59 @@ again:
        s->max_fileno = -1;
        gettimeofday(now, NULL);
        sched_preselect(s);
-       if (list_empty(&pre_select_list) && list_empty(&post_select_list))
+       if (list_empty(&s->pre_select_list) && list_empty(&s->post_select_list))
                return 0;
-       ret = s->select_function(s->max_fileno + 1, &s->rfds, &s->wfds,
-               &s->select_timeout);
-       if (ret < 0)
-               return ret;
-       if (ret == 0) {
-               /*
-                * APUE: Be careful not to check the descriptor sets on return
-                * unless the return value is greater than zero. The return
-                * state of the descriptor sets is implementation dependent if
-                * either a signal is caught or the timer expires.
-                */
+       if (!timeout_is_zero(s)) {
+               ret = s->select_function(s->max_fileno + 1, &s->rfds, &s->wfds,
+                       &s->select_timeout);
+               if (ret < 0)
+                       return ret;
+               if (ret == 0) {
+                       /*
+                        * APUE: Be careful not to check the descriptor sets on return
+                        * unless the return value is greater than zero. The return
+                        * state of the descriptor sets is implementation dependent if
+                        * either a signal is caught or the timer expires.
+                        */
+                       FD_ZERO(&s->rfds);
+                       FD_ZERO(&s->wfds);
+               }
+               gettimeofday(now, NULL);
+       } else {
                FD_ZERO(&s->rfds);
                FD_ZERO(&s->wfds);
        }
-       gettimeofday(now, NULL);
        sched_post_select(s);
-       if (list_empty(&pre_select_list) && list_empty(&post_select_list))
+       if (list_empty(&s->pre_select_list) && list_empty(&s->post_select_list))
                return 0;
        goto again;
 }
 
-/*
- * Initialize the paraslash scheduler.
- */
-static void init_sched(void)
-{
-       PARA_INFO_LOG("initializing scheduler\n");
-       INIT_LIST_HEAD(&pre_select_list);
-       INIT_LIST_HEAD(&post_select_list);
-       initialized = 1;
-}
-
 /**
  * Add a task to the scheduler.
  *
- * \param t the task to add
+ * \param t The task to add.
+ * \param s The scheduler instance to add the task to.
  *
  * If the pre_select pointer of \a t is not \p NULL, it is added to
  * the pre_select list of the scheduler. Same goes for post_select.
  *
  * \sa task::pre_select, task::post_select
  */
-void register_task(struct task *t)
+void register_task(struct sched *s, struct task *t)
 {
-       if (!initialized)
-               init_sched();
        PARA_INFO_LOG("registering %s (%p)\n", t->status, t);
+       if (!s->pre_select_list.next)
+               INIT_LIST_HEAD(&s->pre_select_list);
+       if (!s->post_select_list.next)
+               INIT_LIST_HEAD(&s->post_select_list);
        if (t->pre_select) {
                PARA_DEBUG_LOG("pre_select: %p\n", &t->pre_select);
-               list_add_tail(&t->pre_select_node, &pre_select_list);
+               list_add_tail(&t->pre_select_node, &s->pre_select_list);
        }
        if (t->post_select) {
                PARA_DEBUG_LOG("post_select: %p\n", &t->post_select);
-               list_add_tail(&t->post_select_node, &post_select_list);
+               list_add_tail(&t->post_select_node, &s->post_select_list);
        }
 }
 
@@ -195,21 +197,18 @@ void register_task(struct task *t)
  * \a pre_select_list and the \a post_select_list are empty. This function
  * must be called from the post_select (rather than the pre_select) method.
  */
-void sched_shutdown(void)
+void sched_shutdown(struct sched *s)
 {
        struct task *t, *tmp;
 
-       if (!initialized)
-               return;
-       list_for_each_entry_safe(t, tmp, &pre_select_list, pre_select_node) {
+       list_for_each_entry_safe(t, tmp, &s->pre_select_list, pre_select_node) {
                t->error = -E_SCHED_SHUTDOWN;
                unregister_task(t);
        }
-       list_for_each_entry_safe(t, tmp, &post_select_list, post_select_node) {
+       list_for_each_entry_safe(t, tmp, &s->post_select_list, post_select_node) {
                t->error = -E_SCHED_SHUTDOWN;
                unregister_task(t);
        }
-       initialized = 0;
 }
 
 /**
@@ -220,20 +219,18 @@ void sched_shutdown(void)
  * Each entry of the list contains an identifier which is simply a hex number.
  * The result is dynamically allocated and must be freed by the caller.
  */
-char *get_task_list(void)
+char *get_task_list(struct sched *s)
 {
        struct task *t, *tmp;
        char *msg = NULL;
 
-       if (!initialized)
-               return NULL;
-       list_for_each_entry_safe(t, tmp, &pre_select_list, pre_select_node) {
+       list_for_each_entry_safe(t, tmp, &s->pre_select_list, pre_select_node) {
                char *tmp_msg;
                tmp_msg = make_message("%s%p\tpre\t%s\n", msg? msg : "", t, t->status);
                free(msg);
                msg = tmp_msg;
        }
-       list_for_each_entry_safe(t, tmp, &post_select_list, post_select_node) {
+       list_for_each_entry_safe(t, tmp, &s->post_select_list, post_select_node) {
                char *tmp_msg;
 //             if (t->pre_select)
 //                     continue;
@@ -297,17 +294,19 @@ void sched_request_timeout_ms(long unsigned ms, struct sched *s)
  * \param barrier Absolute time before select() should return.
  * \param s Pointer to the scheduler struct.
  *
* If \a barrier is in the past, this function does nothing.
  *
  * \sa sched_request_barrier_or_min_delay().
  */
-void sched_request_barrier(struct timeval *barrier, struct sched *s)
+int sched_request_barrier(struct timeval *barrier, struct sched *s)
 {
        struct timeval diff;
 
        if (tv_diff(now, barrier, &diff) > 0)
-               return;
+               return 0;
        sched_request_timeout(&diff, s);
+       return 1;
 }
 
 /**
@@ -316,15 +315,19 @@ void sched_request_barrier(struct timeval *barrier, struct sched *s)
  * \param barrier Absolute time before select() should return.
  * \param s Pointer to the scheduler struct.
  *
* If \a barrier is in the past, this function requests a minimal timeout.
  *
  * \sa sched_min_delay(), sched_request_barrier().
  */
-void sched_request_barrier_or_min_delay(struct timeval *barrier, struct sched *s)
+int sched_request_barrier_or_min_delay(struct timeval *barrier, struct sched *s)
 {
        struct timeval diff;
 
-       if (tv_diff(now, barrier, &diff) > 0)
-               return sched_min_delay(s);
+       if (tv_diff(now, barrier, &diff) > 0) {
+               sched_min_delay(s);
+               return 0;
+       }
        sched_request_timeout(&diff, s);
+       return 1;
 }
@@ -297,17 +294,19 @@ void sched_request_timeout_ms(long unsigned ms, struct sched *s)
  * \param s Pointer to the scheduler struct.
  *
- * If \a barrier is in the past, this function does nothing.
+ * \return If \a barrier is in the past, this function does nothing and returns
+ * zero. Otherwise it returns one.
  *
  * \sa sched_request_barrier_or_min_delay().
@@ -316,15 +315,19 @@ void sched_request_barrier(struct timeval *barrier, struct sched *s)
  * \param s Pointer to the scheduler struct.
  *
- * If \a barrier is in the past, this function requests a minimal timeout.
+ * \return If \a barrier is in the past, this function requests a minimal
+ * timeout and returns zero. Otherwise it returns one.
  *
  * \sa sched_min_delay(), sched_request_barrier().